/**
 * @file fstv.c
 *
 * @brief Uses the fist and second order TV method to inpaint an image.
 *
 * The algorithm is based on the paper by K. Papafitsoros, C. Schoenleib and B. Sengul. The objective is to obtain the following minimisation,
 * @f[
 *  u_r=\underset{u\in \text{BH}(\Omega)}{\operatorname{argmin}}\int_{\Omega \backslash D} (u-u_0)^2\, dx + \alpha \text{TV}(u)+\beta\text{TV}^2(u).
 * @f]
 * For more information consult the paper.
 *
 * @author Bati Sengul
 *
 * @version 1.3
 *
 * @section LICENSE
 *
 *    Copyright (c) 2012, Bati Sengul
 *
 *	 This program is free software: you can use, modify and/or
 *	 redistribute it under the terms of the simplified BSD License. You
 *	 should have received a copy of this license along this program. If
 *	 not, see <http://opensource.org/licenses/BSD-3-Clause>.
 *
 * @todo Implement TIFF and alpha handling.
 * @todo Improve efficiency by not calling In, Ic so much.
 */ 


#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <png.h>
#include <complex.h>
#include <math.h>
#include <fftw3.h>
#include "io_png.h"

/**
 *
 * @brief Implementations of a malloc that quits when the return of malloc is null.
 *
 */
inline void* xmalloc(size_t size){
		void* mem = malloc(size);
		if( mem == NULL){
				perror("Memory allocation error");
				exit(EXIT_FAILURE);
		}
		return mem;
}


#define DPAD 0.01
/**
 * @brief The threashold for the Frobenius distance to reach before the algorithm stops.
 */ 
#define THRESHOLD 0.0000000064
/**
 * @brief The maximum number of iterations the algorithm will run for.
 */ 
#define MAXITER 2000
#define MIN(x,y) x>=y ? y:x

inline float MAX(float x, float y){
		return x<=y ? y:x;
}


/**
 * @brief Height of the image.
 */ 
int height;
/**
 * @brief Width of the image.
 */ 
int width;
/**
 * @brief Parameter governing the first order TV.
 */ 
float a;
/**
 * @brief Parameter governing the second order TV.
 */ 
float b;
/**
 * @brief The parameter \f$\lambda_1\f$ as described in the paper.
 */
float l1;
/**
 * @brief The parameter \f$\lambda_2\f$ as described in the paper.
 */
float l2;
/**
 * @brief The parameter \f$\lambda_0\f$ as described in the paper.
 */
float l0;


/*
 * We will be using a/l1 and b/l2 a lot. We store them in al and bl respectively.
 */
float al, bl;

fftw_plan p[3], pinv[3];
fftw_complex *out[3];
double *in[3];

/**
 * @brief Calculates the position on a torus.
 *
 * @param i The y coordinate
 * @param j The x coordinate
 * @return The position in an array of the coordinate (i,j)
 */ 
inline int In(int i, int j){
		if(i < 0)
				i += height;
		else if( i >= height)
				i = i - height;
		if(j < 0)
				j += width;  
		else if( j >= width)
				j = j - width;
		return i*width + j;
}
/**
 * @brief Calculates the position on a torus.
 *
 * @param i The y coordinate
 * @param j The x coordinate
 * @param c The channel number
 * @return The position in an array of the coordinate (i,j) on channel c
 */ 
inline int Ic(int i, int j, int c){
		if(i < 0)
				i += height;  
		else if( i >= height)
				i = i - height;
		if(j < 0)
				j += width;  
		else if( j >= width)
				j = j - width;
		return c*height*width + i*width + j;
}

/**
 * @brief The forward derivative in the x coordinate.
 *
 * @param x The array to take the derivate from
 * @param i The x coordinate of the derivative
 * @param j The y coordinate of the derivative
 * @return The forward x derivative at position (i,j)
 */ 
inline float fdx(const float *x, int i, int j){
		return x[In(i,j+1)] - x[In(i,j)];
}
/**
 * @brief The forward derivative in the y coordinate.
 *
 * @param x The array to take the derivate from
 * @param i The x coordinate of the derivative
 * @param j The y coordinate of the derivative
 * @return The forward y derivative at position (i,j)
 */ 
inline float fdy(const float *x, int i, int j){
		return x[In(i+1,j)] - x[In(i,j)];
}
/**
 * @brief The backward derivative in the x coordinate.
 *
 * @param x The array to take the derivate from
 * @param i The x coordinate of the derivative
 * @param j The y coordinate of the derivative
 * @return The backward x derivative at position (i,j)
 */ 
inline float bdx(const float *x, int i,int j){
		return x[In(i,j)] - x[In(i,j-1)];
}
/**
 * @brief The backward derivative in the y coordinate.
 *
 * @param x The array to take the derivate from
 * @param i The x coordinate of the derivative
 * @param j The y coordinate of the derivative
 * @return The backward y derivative at position (i,j)
 */ 
inline float bdy(const float *x, int i, int j){
		return x[In(i,j)] - x[In(i-1,j)];
}
/**
 * @brief The second derivative in the x coordinate.
 *
 * @param x The array to take the derivate from
 * @param i The x coordinate of the derivative
 * @param j The y coordinate of the derivative
 * @return The second derivative in the x direction at position (i,j)
 */ 
inline float dxx(const float *x, int i,int j){
		return x[In(i,j+1)] + x[In(i,j-1)] - 2*x[In(i,j)];
}
/**
 * @brief The second derivative in the y coordinate.
 *
 * @param x The array to take the derivate from
 * @param i The x coordinate of the derivative
 * @param j The y coordinate of the derivative
 * @return The second derivative in the y direction at position (i,j)
 */ 
inline float dyy(const float *x, int i, int j){
		return x[In(i+1,j)] + x[In(i-1,j)] - 2*x[In(i,j)];
}
/**
 * @brief The forward mixed derivative.
 *
 * @param x The array to take the derivate from
 * @param i The x coordinate of the derivative
 * @param j The y coordinate of the derivative
 * @return The forward mixed derivative at position (i,j)
 */ 
inline float dxy(const float *x, int i, int j){
		return x[In(i,j)] + x[In(i+1,j+1)] - x[In(i+1,j)] - x[In(i,j+1)];
}
/**
 * @brief The backward mixed derivative.
 *
 * @param x The array to take the derivate from
 * @param i The x coordinate of the derivative
 * @param j The y coordinate of the derivative
 * @return The backward mixed derivative at position (i,j)
 */ 
inline float bdxy(const float *x, int i, int j){
		return x[In(i,j)] + x[In(i-1,j-1)] - x[In(i-1,j)] - x[In(i,j-1)];
}

/**
 * @brief Calculates the Frobenius distance between two arrays.
 *
 * Calculates a Frobenius like distance to test convergence to the solution. The following formula
 * @f[
 * d(A,B)=\frac{\sum_{i,j=1}^n (a_{i,j}-b_{i,j})^2}{\sum_{i=1}^n b_{i,j}^2}.
 * @f]
 * is applied to each channel. The maximum over the channels is returned.
 *
 * @param A The first image
 * @param B The second image
 * @return The distance between the two images
 */

inline float froDist(const float *A, const float *B){
		float top[3] = {0,0,0};
		float bot[3] = {0,0,0};
		int i, j, c;
#pragma omp parallel for \
		shared(top,bot,A,B) \
		private(c,i,j)
		for(c = 0; c < 3; c++){
				for(i = 0; i < height; i++){
						for(j = 0; j < width; j++){
								top[c] += (A[Ic(i,j,c)] - B[Ic(i,j,c)]) * (A[Ic(i,j,c)] - B[Ic(i,j,c)]);
								bot[c] += B[Ic(i,j,c)] * B[Ic(i,j,c)];
						}
				}
		}
		return MAX(MAX(top[0]/bot[0], top[1]/bot[1]),top[2]/bot[2]);

}


void printHelp(){
		printf("Combined first and second order TV inpainting using split Bregman, K. Papafitsoros, C. Schoenleib, B. Sengul, 2012\n\n");
		printf("Syntax: fstv [image] [mask] [output] [options]\n\n");
		printf("[image] must be a PNG image which will be inpainted\n");
		printf("[mask] must be a PNG image with the white region denoting the inpainting domain\n");
		printf("[output] name of the inpainted PNG image\n\n");
		printf("[options] are optional and will be interpreted in the following order:\n");
		printf("  alpha               [default = 0.001]\n");
		printf("  beta                [default = 0.001]\n");
}



/**
 * @brief Sets the mask from an image.
 *
 * The mask will be obtained by looking at the area that has RGB values (254,254,254) or higher.
 *
 * @param mask The mask that will be extracted. Make sure this has the same size allocation as the second argument
 * @param im The image from which the mask will be extracted
 * */
void setMask(float *mask, const float *im){
		int i, j, c;
		int val;
		for(i = 0; i < height; i++){
				for(j = 0; j < width; j++){
						if(im[Ic(i,j,0)] > 253 && im[Ic(i,j,1)] > 253  && im[Ic(i,j,2)] > 253)
								val = 0;
						else
								val = 1;
						for(c = 0; c < 3; c++){
								mask[Ic(i,j,c)] = val;
						}
				}
		}
}


/**
 * @brief Sets the value of an array to a given value.
 *
 * @param A The array to be written on
 * @param value The value to write to the array
 * @param size The size of the array
 */

void setValue(float *A, float value, int size){
		int i;
		for(i = 0; i < size; i++)
				A[i] = value;
}

/**
 * @brief Computes the denominator of the Fourier exponent.
 *
 * @return The denominator of the Fourier exponent. A dynamic memory has been allocated.
 */ 
fftw_complex* computeDenom(){		
		int i;
		fftw_complex *ret;
		ret =  (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*(width/2+1)*height);
		for(i = 0; i < height*(width/2+1); i++)
				ret[i] = l0;

		/* dxx Matrix */
		memset(in[1], 0, sizeof(double)*width*height);
		in[1][In(0,0)] = -2;
		in[1][In(0,1)] = 1;
		in[1][In(0,width-1)] = 1;

		fftw_execute(p[1]);

		for(i = 0; i < height*(width/2+1); i++)
				ret[i] -= l1*out[1][i];


		/* dyy Matrix */
		memset(in[2], 0, sizeof(double)*width*height);
		in[2][In(0,0)] = -2;
		in[2][In(1,0)] = 1;
		in[2][In(height-1,0)] = 1;

		fftw_execute(p[2]);

		for(i = 0; i < height*(width/2+1); i++)
				ret[i] -= l1*out[2][i];


		/* d4x Matrix */
		memset(in[0], 0, sizeof(double)*width*height);
		in[0][In(0,0)] = 6;
		in[0][In(0,1)] = -4;
		in[0][In(0,2)] = 1;
		in[0][In(0, width-2)] = 1;
		in[0][In(0, width-1)] = -4;

		fftw_execute(p[0]);

		for(i = 0; i < height*(width/2+1); i++)
				ret[i] += l2*out[0][i];



		/* d4y Matrix */
		memset(in[1], 0, sizeof(double)*width*height);
		in[1][In(0,0)] = 6;
		in[1][In(1,0)] = -4;
		in[1][In(2,0)] = 1;
		in[1][In(height-2,0)] = 1;
		in[1][In(height-1,0)] = -4;

		fftw_execute(p[1]);

		for(i = 0; i < height*(width/2+1); i++)
				ret[i] += l2*out[1][i];


		/* d2x2y Matrix */
		memset(in[1], 0, sizeof(double)*width*height);
		in[1][In(0,0)] = 4;
		in[1][In(1,0)] = -2;
		in[1][In(0,1)] = -2;
		in[1][In(1,1)] = 1;
		in[1][In(height-1,0)] = -2;
		in[1][In(height-1,1)] = 1;
		in[1][In(0,width-1)] = -2;
		in[1][In(1,width-1)] = 1;
		in[1][In(height-1,width-1)] = 1;

		fftw_execute(p[1]);

		for(i = 0; i < height*(width/2+1); i++)
				ret[i] += 2*l2*out[1][i];



		return ret;
}

/**
 * @brief Runs the first and second order TV algorithm.
 *
 * The algorithm will run until the Frobenius distance of the consecutive images falls below a threshold or the maximum number of iterations have been reached.
 *
 * @param image The image from which the inpainting will be based off.
 * @param mask The mask which indicates the inpainting domain.
 * @return The inpainted image for which a memory has been dynamically allocated.
 * @sa THRESHOLD MAXITER 
 */

float* fsTV(const float *image, const float *mask){
		int i, j, c;

		/*
		 * The variables below are as they appear on the paper, with the following exception.
		 *
		 * tu denotes \tilde{u} from Subproblem 2
		 *
		 * s denotes the matrix s from Subproblem 3
		 * ts denotes the matrix t from Subproblem 4
		 *
		 */
		float *u, *tu, *u_old;
		float *v1, *v2, *w11, *w22, *w12;
		float *b0, *b1, *b2, *b11, *b22, *b12;
		float *s, *ts;

		u = (float*) xmalloc(sizeof(float)*width*height*3);
		u_old = (float*) xmalloc(sizeof(float)*width*height*3);
		tu = (float*) xmalloc(sizeof(float)*width*height*3);
		v1 = (float*) xmalloc(sizeof(float)*width*height*3);
		v2 = (float*) xmalloc(sizeof(float)*width*height*3);
		w11 = (float*) xmalloc(sizeof(float)*width*height*3);
		w22 = (float*) xmalloc(sizeof(float)*width*height*3);
		w12 = (float*) xmalloc(sizeof(float)*width*height*3);
		b0 = (float*) xmalloc(sizeof(float)*width*height*3);
		b1 = (float*) xmalloc(sizeof(float)*width*height*3);
		b2 = (float*) xmalloc(sizeof(float)*width*height*3);
		b11 = (float*) xmalloc(sizeof(float)*width*height*3);
		b22 = (float*) xmalloc(sizeof(float)*width*height*3);
		b12 = (float*) xmalloc(sizeof(float)*width*height*3);
		s = (float*) xmalloc(sizeof(float)*width*height);
		ts = (float*) xmalloc(sizeof(float)*width*height);

		setValue(b0, 0.01, width*height*3);
		setValue(b1, 0.01, width*height*3);
		setValue(b2, 0.01, width*height*3);
		setValue(b11, 0.01, width*height*3);
		setValue(b22, 0.01, width*height*3);
		setValue(b12, 0.01, width*height*3);

		/* Pointers to the channels */
		float *tu_g, *tu_b;
		float *b1_g, *b1_b, *b2_g, *b2_b, *b11_g, *b11_b, *b22_g, *b22_b, *b12_g, *b12_b;
		b1_g = b1 + width*height;
		b2_g = b2 + width*height;
		b11_g = b11 + width*height;
		b22_g = b22 + width*height;
		b12_g = b12 + width*height;
		/* u_g = u + width*height; */
		tu_g = tu + width*height;

		b1_b = b1_g + width*height;
		b2_b = b2_g + width*height;
		b11_b = b11_g + width*height;
		b22_b = b22_g + width*height;
		b12_b = b12_g + width*height;
		/* u_b = u_g + width*height; */
		tu_b = tu_g + width*height;

		memcpy(u, image, sizeof(float)*width*height*3);

		memcpy(tu, u, sizeof(float)*width*height*3);

		for(c = 0; c < 3; c++){
				for(i = 0; i < height; i++){
						for(j = 0; j < width; j++){
								v1[Ic(i,j,c)] = fdx((u + c*width*height), i, j); 
								v2[Ic(i,j,c)] = fdy((u + c*width*height), i, j);
								w11[Ic(i,j,c)] = dxx((u + c*width*height), i, j); 
								w22[Ic(i,j,c)] = dyy((u + c*width*height), i, j); 
								w12[Ic(i,j,c)] = dxy((u + c*width*height), i, j); 
						}
				}
		}



		/* Fourier tools */
		for(i = 0; i < 3; i++){
				in[i] = (double*) xmalloc(sizeof(double)*width*height);
				out[i] =  (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*(width/2+1)*height);
				p[i] = fftw_plan_dft_r2c_2d(height, width, in[i], out[i], FFTW_MEASURE);
				pinv[i] = fftw_plan_dft_c2r_2d(height, width, out[i], in[i], FFTW_MEASURE);
		}

		fftw_complex *D;
		D = computeDenom();




		/*
		 *
		 * 
		 * Start of iteration
		 *
		 *
		 */
		int it = 0;
		int chan;
		do {
				memcpy(u_old,u,sizeof(float)*width*height*3);

#pragma omp parallel for \
				shared(u,mask,image,b0,tu,l0,l1,l2,b1,b2,b11,b12,b22,v1,v2,w11,w12,w22,D,in,out) \
				private(c,i,j,chan) 
				for(c = 0; c < 3; c++){
						chan = c*width*height;
						/* Subproblem 1 */
						for(i = 0; i < height; i++){
								for(j = 0; j < width; j++) 
										u[i*width+j+chan] = (2/(2+l0)) * mask[i*width+j+chan] * image[i*width+j+chan] + (b0[i*width+j+chan] + tu[i*width+j+chan]) * (l0 + 2*(1-mask[i*width+j+chan])) / (l0+2);
						}

						/* Subproblem 2 using the Fourier method */
						for(i = 0; i < height; i++){
								for(j = 0; j < width; j++){
										in[c][i*width+j] = l0*(u[i*width+j+chan]- b0[i*width+j+chan]) 
												+ l1*( bdx((b1+chan),i,j) - bdx((v1+chan),i,j) + bdy((b2+chan),i,j) - bdy((v2+chan),i,j) )
												- l2*( dxx((b11+chan),i,j) + dyy((b22+chan),i,j) + 2*bdxy((b12+chan),i,j) )
												+ l2*( dxx((w11+chan),i,j) + dyy((w22+chan),i,j) + 2*bdxy((w12+chan),i,j) );
								}
						}

						fftw_execute(p[c]);

						for(i = 0; i < height*(width/2+1); i++)
								out[c][i] /= D[i];

						fftw_execute(pinv[c]);


						for(i = 0; i < height; i++){
								for(j = 0; j < width; j++)
										tu[i*width+j+chan] = in[c][i*width+j] / (float) (height*width);
						}

				}

				/* Calculating mixing matricies for Subproblems 3,4 */
#pragma omp parallel for \
				shared(b0,tu,b1,b2,b11,b12,b22,b1_g,b2_g,b11_g,b12_g,b22_g,b1_b,b2_b,b11_b,b12_b,b22_b,s,ts) \
				private(c,i,j) 
				for(i = 0; i < height; i++){
						for(j = 0; j < width; j++){
								if(a != 0){
										s[i*width+j] =  sqrt ( (fdx(tu,i,j)+b1[i*width+j]) * (fdx(tu,i,j)+b1[i*width+j]) 
														+ (fdy(tu,i,j)+b2[i*width+j]) * (fdy(tu,i,j)+b2[i*width+j])
														+ (fdx(tu_g,i,j)+b1_g[i*width+j]) * (fdx(tu_g,i,j)+b1_g[i*width+j]) 
														+ (fdy(tu_g,i,j)+b2_g[i*width+j]) * (fdy(tu_g,i,j)+b2_g[i*width+j])
														+ (fdx(tu_b,i,j)+b1_b[i*width+j]) * (fdx(tu_b,i,j)+b1_b[i*width+j]) 
														+ (fdy(tu_b,i,j)+b2_b[i*width+j]) * (fdy(tu_b,i,j)+b2_b[i*width+j])  );
								}
								if(b != 0){
										ts[i*width+j] =  sqrt ( (dxx(tu,i,j)+b11[i*width+j]) * (dxx(tu,i,j)+b11[i*width+j]) 
														+ (dyy(tu,i,j)+b22[i*width+j]) * (dyy(tu,i,j)+b22[i*width+j])
														+ (dxx(tu_g,i,j)+b11_g[i*width+j]) * (dxx(tu_g,i,j)+b11_g[i*width+j]) 
														+ (dyy(tu_g,i,j)+b22_g[i*width+j]) * (dyy(tu_g,i,j)+b22_g[i*width+j])
														+ (dxx(tu_b,i,j)+b11_b[i*width+j]) * (dxx(tu_b,i,j)+b11_b[i*width+j]) 
														+ (dyy(tu_b,i,j)+b22_b[i*width+j]) * (dyy(tu_b,i,j)+b22_b[i*width+j]) 
														+ 2* (dxy(tu,i,j)+b12[i*width+j]) * (dxy(tu,i,j)+b12[i*width+j])
														+ 2* (dxy(tu_g,i,j)+b12_g[i*width+j]) * (dxy(tu_g,i,j)+b12_g[i*width+j])
														+ 2* (dxy(tu_b,i,j)+b12_b[i*width+j]) * (dxy(tu_b,i,j)+b12_b[i*width+j])	);
								}

						}
				}
#pragma omp parallel for \
				shared(u,tu,b0,b1,b2,b11,b12,b22,v1,v2,w11,w12,w22,s,ts) \
				private(c,i,j,chan) 
				for(c = 0; c < 3; c++){
						chan = c*width*height;
						/* Subproblem 3 */
						if(a != 0){
								for(i = 0; i < height; i++){
										for(j = 0; j < width; j++){
												v1[i*width+j+chan] = MAX(s[i*width+j] - al, 0) * (fdx((tu+chan), i, j)+b1[i*width+j+chan])/MAX(s[i*width+j], DPAD);
										}
								}
								for(i = 0; i < height; i++){
										for(j = 0; j < width; j++){
												v2[i*width+j+chan] = MAX(s[i*width+j] - al, 0) * (fdy((tu+chan), i, j)+b2[i*width+j+chan])/MAX(s[i*width+j], DPAD);
										}
								}
						}

						/* Subproblem 4 */
						if(b !=0 ){
								for(i = 0; i < height; i++){
										for(j = 0; j < width; j++){
												w11[i*width+j+chan] = MAX(ts[i*width+j] - bl, 0) * (dxx((tu+chan), i, j)+b11[i*width+j+chan])/MAX(ts[i*width+j], DPAD);
										}
								}
								for(i = 0; i < height; i++){
										for(j = 0; j < width; j++){
												w12[i*width+j+chan] = MAX(ts[i*width+j] - bl, 0) * (dxy((tu+chan), i, j)+b12[i*width+j+chan])/MAX(ts[i*width+j], DPAD);
										}
								}
								for(i = 0; i < height; i++){
										for(j = 0; j < width; j++){
												w22[i*width+j+chan] = MAX(ts[i*width+j] - bl, 0) * (dyy((tu+chan), i, j)+b22[i*width+j+chan])/MAX(ts[i*width+j], DPAD);
										}
								}
						}

						/* Updates on b0, b1, b2 */
						for(i = 0; i < height; i++){
								for(j = 0; j < width; j++){
										b0[i*width+j+chan] += tu[i*width+j+chan] - u[i*width+j+chan];
								}
						}
						for(i = 0; i < height; i++){
								for(j = 0; j < width; j++){
										b1[i*width+j+chan] += fdx((tu+chan),i,j) - v1[i*width+j+chan];
								}
						}
						for(i = 0; i < height; i++){
								for(j = 0; j < width; j++){
										b2[i*width+j+chan] += fdy((tu+chan),i,j) - v2[i*width+j+chan];
								}
						}
						for(i = 0; i < height; i++){
								for(j = 0; j < width; j++){
										b11[i*width+j+chan] += dxx((tu+chan),i,j) - w11[i*width+j+chan];
								}
						}
						for(i = 0; i < height; i++){
								for(j = 0; j < width; j++){
										b22[i*width+j+chan] += dyy((tu+chan),i,j) - w22[i*width+j+chan];
								}
						}
						for(i = 0; i < height; i++){
								for(j = 0; j < width; j++){
										b12[i*width+j+chan] += dxy((tu+chan),i,j) - w12[i*width+j+chan];
								}
						}
				}
				it++;

		} while(it < MAXITER && froDist(u,u_old)> THRESHOLD);
		printf("Iterations: %i     MaxIterations: %i\nFrobenius Distance: %e     Threshold: %e\n",it,MAXITER, froDist(u,u_old), THRESHOLD);
		if(it == MAXITER)
				printf("Warning: Maximum iterations reached!\n");


		/* Cleanup */
		for(i = 0; i < 3; i++){
				fftw_destroy_plan(p[i]);
				fftw_destroy_plan(pinv[i]);
				free(in[i]);
				free(out[i]);
		}

		free(tu);
		free(u_old);
		free(v1);
		free(v2);
		free(w11);
		free(w22);
		free(w12);
		free(s);
		free(ts);
		free(b0);
		free(b1);
		free(b2);
		free(b11);
		free(b22);
		free(b12);
		fftw_free(D);

		return u;

}

int main(int argc, char **argv){
		if(argc > 1 && (!strcmp(argv[1],"--help") || !strcmp(argv[1],"-h"))){
				printHelp();
				return 0;
		}
		if(argc < 4){
				printf("Too few arguments.\n\n");
				printf("Syntax: fstv [image] [mask] [output] [options]\n");
				printf("  try --help for more\n");
				return 0;
		}

		float *im = NULL;
		size_t swidth, sheight;
		im = io_png_read_f32_rgb(argv[1], &swidth, &sheight);
		if(im == NULL){
				printf("Are you sure %s is a png image?\n", argv[1]);
				return 0;
		}

		width = (int) swidth;
		height = (int) sheight;

		float *m;
		m = io_png_read_f32_rgb(argv[2], &swidth, &sheight);
		if(m == NULL){
				printf("Are you sure %s is a png image?\n", argv[1]);
				free(im);
				return 0;
		}
		if(width != (int) swidth || height != (int) sheight){
				printf("Dimensions of %s and %s don't agree!\n",argv[1],argv[2]);
				free(im);
				free(m);
				return 0;
		}
		setMask(m,m);
		int i;
		for(i = 0; i < width*height*3; i++)
				im[i] /= 255;

		a = 0.001;
		b = 0.001;

		if (argc > 4){
				a = (float) atof(argv[4]);
		}
		if(a < 0 )
				a = 0.001;
		if (argc > 5){
				b = (float) atof(argv[5]);
		}
		if(b < 0 )
				b = 0.001;

		if(a == 0 && b == 0){
				printf("Here is your original image back!\n");
				io_png_write_f32(argv[3], im, swidth, sheight, 3);
				free(im);
				free(m);
				return 0;
		}
		l1 = 10*a;
		l2 = 10*b;
		l0 = MAX(a,b);

		if(a != 0)
				al = a/l1;
		if(b != 0)
				bl = b/l2;
		printf("Running algorithm with parameters:\nalpha:   %f\nbeta:    %f\nlambda0: %f\nlambda1: %f\nlambda2: %f\n\n",a,b,l0,l1,l2);

		float *res = fsTV(im,m);
		for(i = 0; i < height*width*3; i++)
				res[i] *= 255;
		io_png_write_f32(argv[3], res, swidth, sheight, 3);

		free(im);
		free(m);
		free(res);
		return 0;
}

/**
 * @mainpage First and Second Order TV
 * @section intro Introduction
 * This code was written for IPOL and is based on the paper "Combined first and second order total variation inpainting using split Bregman, K. Papafitsoros, B. Sengul, C. Schoenlieb".
 * The program will solve the following minimisation problem:
 * @f[
 *  u_r=\underset{u\in \text{BH}(\Omega)}{\operatorname{argmin}}\int_{\Omega \backslash D} (u-u_0)^2\, dx + \alpha \text{TV}(u)+\beta\text{TV}^2(u).
 * @f]
 *
 * The code uses the first and second order total variation inpainting method to inpaint, for more information see @ref usage "usage".
 *
 * @section install Installation
 * Run
 * @code make clean; make @endcode
 * to compile the code. This requires the following libraries:
 * @li fftw3
 * @li libpng
 * @li openmp
 * 
 * the latter can be commented out in the Makefile.
 *
 * The testing was done on Mac OS X 10.6.4 with gcc version 4.2.1 and 4.7
 *
 * @section usage Usage
 * To print out all the options and explanations use
 * @code fstv --help @endcode
 * The program takes inputs in the following way:
 * @code fstv [image] [mask] [output] [options] @endcode
 *
 *		@li @code [image]   @endcode PNG image to be inpainted 
 *		@li @code [mask]    @endcode PNG image with white region specifying the inpainting domain
 *		@li @code [output]  @endcode	filename of the inpainted PNG
 *		@li @code [options] @endcode @code alpha [default 0.001] beta [default 0.001]@endcode
 *
 * The [image] and [mask] images must have the same dimensions.
 *
 * The software comes with two test images and the inpainting can be done as follows:
 * @code fstv resting.png resting.png resting_inpainted.png @endcode
 * For the second picture we will run it with parameters alpha=0.0001 and beta=0.007 as follows:
 * @code fstv sotirisbig.png sotirisbig.png sotirisbig_inpainted.png 0.0001 0.007 @endcode
 * 
 * @section licence Licence
 * @subsection fstv fstv.c
 *
 *    Copyright (c) 2012, Bati Sengul <b.sengul@maths.cam.ac.uk>
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the simplified BSD License. You
 * should have received a copy of this license along this program. If
 * not, see <http://opensource.org/licenses/BSD-3-Clause>.
 *
 *  @subsection io_png io_png.c
 *  
 * Copyright (c) 2010-2011, Nicolas Limare <nicolas.limare@cmla.ens-cachan.fr>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under, at your option, the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version, or
 * the terms of the simplified BSD license.
 *
 * You should have received a copy of these licenses along this
 * program. If not, see <http://www.gnu.org/licenses/> and
* <http://www.opensource.org/licenses/bsd-license.html>.
*
*/ 
